@@ -125,10 +125,15 @@ already ~a threads running, disabling 'signalfd' support")
 
 (define (handle-SIGINT)
   "Handle SIGINT by stopping the Shepherd, which means rebooting if we're PID 1."
-  (catch 'quit
-    (lambda ()
-      (stop root-service))
-    quit-exception-handler))
+  ;; Since 'stop' is synchronous and may block until SIGCHLD has been received
+  ;; for the process it's waiting for, call it in a separate fiber so that
+  ;; signals are still being processed in the meantime.
+  (spawn-fiber
+   (lambda ()
+     (catch 'quit
+       (lambda ()
+         (stop root-service))
+       quit-exception-handler))))
 
 (define (signal-handler signal)
   "Return the signal handler for SIGNAL."
@@ -137,11 +142,7 @@ already ~a threads running, disabling 'signalfd' support")
         ((= signal SIGINT)
          (lambda _ (handle-SIGINT)))
         ((memv signal (list SIGTERM SIGHUP))
-         (lambda _
-           (catch 'quit
-             (lambda ()
-               (stop root-service))
-             quit-exception-handler)))
+         (lambda _ (handle-SIGINT)))
         (else
          (const #f))))
 
@@ -415,12 +416,29 @@ already ~a threads running, disabling 'signalfd' support")
        (lambda ()
          (catch 'quit
            (lambda ()
-             (run-daemon #:socket-file socket-file
-                         #:config-file config-file
-                         #:pid-file pid-file
-                         #:signal-port signal-port
-                         #:poll-services? poll-services?
-                         #:persistency persistency))
+             (with-process-monitor
+
+               ;; Replace the default 'system*' binding with one that
+               ;; cooperates instead of blocking on 'waitpid'.
+               (let ((real-system* system*))
+                 (set! system* spawn-command)
+
+                 ;; Restore 'system*' after fork.
+                 (set! primitive-fork
+                       (let ((real-fork primitive-fork))
+                         (lambda ()
+                           (let ((result (real-fork)))
+                             (when (zero? result)
+                               (set! primitive-fork real-fork)
+                               (set! system* real-system*))
+                             result)))))
+
+               (run-daemon #:socket-file socket-file
+                           #:config-file config-file
+                           #:pid-file pid-file
+                           #:signal-port signal-port
+                           #:poll-services? poll-services?
+                           #:persistency persistency)))
            (case-lambda
              ((key value . _)
               (primitive-exit value))
