@@ -26,7 +26,10 @@
 (define-module (shepherd service)
   #:use-module ((fibers)
                 #:hide (sleep))
+  #:use-module (fibers channels)
+  #:use-module (fibers operations)
   #:use-module (fibers scheduler)
+  #:use-module (fibers timers)
   #:use-module (oop goops)
   #:use-module (srfi srfi-1)
   #:use-module (srfi srfi-9)
@@ -37,8 +40,8 @@
   #:use-module ((ice-9 control) #:select (call/ec))
   #:use-module (ice-9 match)
   #:use-module (ice-9 format)
+  #:use-module (ice-9 vlist)
   #:autoload   (ice-9 ports internal) (port-read-wait-fd)
-  #:autoload   (ice-9 rdelim) (read-line)
   #:autoload   (ice-9 pretty-print) (truncated-print)
   #:use-module (shepherd support)
   #:use-module (shepherd comm)
@@ -76,6 +79,8 @@
             lookup-services
             respawn-service
             handle-SIGCHLD
+            with-process-monitor
+            spawn-command
             %precious-signals
             register-services
             provided-by
@@ -86,6 +91,7 @@
             default-environment-variables
             make-forkexec-constructor
             make-kill-destructor
+            default-process-termination-grace-period
             exec-command
             fork+exec-command
             default-pid-file-timeout
@@ -453,66 +459,64 @@ NEW-SERVICE."
 canonical names for all of the services which have been stopped (including
 transitive dependent services).  This method will print a warning if SERVICE
 is not already running, and will return SERVICE's canonical name in a list."
-  ;; Block asyncs so the SIGCHLD handler doesn't execute concurrently.
-  ;; Notably, that makes sure the handler processes the SIGCHLD for SERVICE's
-  ;; process once we're done; otherwise, it could end up respawning SERVICE.
-  (call-with-blocked-asyncs
-   (lambda ()
-     (if (not (running? service))
-         (begin
-           (local-output (l10n "Service ~a is not running.")
-                         (canonical-name service))
-           (list (canonical-name service)))
-         (if (slot-ref service 'stop-delay?)
-             (begin
-               (slot-set! service 'waiting-for-termination? #t)
-               (local-output (l10n "Service ~a pending to be stopped.")
-                             (canonical-name service))
-               (list (canonical-name service)))
-             (let ((name (canonical-name service))
-                   (stopped-dependents (fold-services (lambda (other acc)
-                                                        (if (and (running? other)
-                                                                 (required-by? service other))
-                                                            (append (stop other) acc)
-                                                            acc))
-                                                      '())))
-               ;; Stop the service itself.
-               (catch #t
-                 (lambda ()
-                   (apply (slot-ref service 'stop)
-                          (service-running-value service)
-                          args))
-                 (lambda (key . args)
-                   ;; Special case: 'root' may quit.
-                   (and (eq? root-service service)
-                        (eq? key 'quit)
-                        (apply quit args))
-                   (caught-error key args)))
+  (if (not (running? service))
+      (begin
+        (local-output (l10n "Service ~a is not running.")
+                      (canonical-name service))
+        (list (canonical-name service)))
+      (if (slot-ref service 'stop-delay?)
+          (begin
+            (slot-set! service 'waiting-for-termination? #t)
+            (local-output (l10n "Service ~a pending to be stopped.")
+                          (canonical-name service))
+            (list (canonical-name service)))
+          (let* ((name (canonical-name service))
+                 (dependents (fold-services (lambda (other lst)
+                                              (if (and (running? other)
+                                                       (required-by? service other))
+                                                  (cons other lst)
+                                                  lst))
+                                            '()))
+                 ;; Note: 'fold-services' introduces a continuation barrier,
+                 ;; which is why we're not using it when calling 'stop'.
+                 (stopped-dependents (append-map stop dependents)))
+            ;; Stop the service itself.
+            (catch #t
+              (lambda ()
+                (let ((running (service-running-value service)))
+                  ;; Mark SERVICE as already stopped to prevent the respawn
+                  ;; machinery from firing upon SIGCHLD.
+                  (slot-set! service 'running #f)
 
-               ;; SERVICE is no longer running.
-               (slot-set! service 'running #f)
+                  (apply (slot-ref service 'stop) running args)))
+              (lambda (key . args)
+                ;; Special case: 'root' may quit.
+                (and (eq? root-service service)
+                     (eq? key 'quit)
+                     (apply quit args))
+                (caught-error key args)))
 
-               ;; Reset the list of respawns.
-               (slot-set! service 'last-respawns '())
+            ;; Reset the list of respawns.
+            (slot-set! service 'last-respawns '())
 
-               ;; Replace the service with its replacement, if it has one
-               (let ((replacement (slot-ref service 'replacement)))
-                 (when replacement
-                   (replace-service service replacement)))
+            ;; Replace the service with its replacement, if it has one
+            (let ((replacement (slot-ref service 'replacement)))
+              (when replacement
+                (replace-service service replacement)))
 
-               ;; Status message.
-               (if (running? service)
-                   (local-output (l10n "Service ~a could not be stopped.")
-                                 name)
-                   (local-output (l10n "Service ~a has been stopped.")
-                                 name))
+            ;; Status message.
+            (if (running? service)
+                (local-output (l10n "Service ~a could not be stopped.")
+                              name)
+                (local-output (l10n "Service ~a has been stopped.")
+                              name))
 
-               (when (transient? service)
-                 (hashq-remove! %services (canonical-name service))
-                 (local-output (l10n "Transient service ~a unregistered.")
-                               (canonical-name service)))
+            (when (transient? service)
+              (hashq-remove! %services (canonical-name service))
+              (local-output (l10n "Transient service ~a unregistered.")
+                            (canonical-name service)))
 
-               (cons name stopped-dependents)))))))
+            (cons name stopped-dependents)))))
 
 ;; Call action THE-ACTION with ARGS.
 (define-method (action (obj <service>) the-action . args)
@@ -553,7 +557,15 @@ is not already running, and will return SERVICE's canonical name in a list."
     ;; it provides generally useful functionality and information.
     (catch #t
       (lambda ()
-        (apply proc (service-running-value obj) args))
+        ;; PROC may return any number of values (e.g., if PROC is
+        ;; 'eval-in-user-module'), including zero values, but callers expect a
+        ;; single value.  Deal with it gracefully.
+        (call-with-values
+            (lambda ()
+              (apply proc (service-running-value obj) args))
+          (case-lambda
+            (() *unspecified*)
+            ((first . rest) first))))
       (lambda (key . args)
         ;; Special case: 'root' may quit.
         (and (eq? root-service obj)
@@ -659,6 +671,20 @@ clients."
 
 
 
+(cond-expand
+ (guile-3.0 #t)
+ (else
+  ;; In Guile 2.2, SRFI-1 'find' is in C and thus introduces a continuation
+  ;; barrier, which is a problem for 'launch-service'.  Provide a Scheme
+  ;; implementation to address that.
+  (define (find pred lst)
+    (let loop ((lst lst))
+      (and (not (null? lst))
+           (let ((head (car lst)))
+             (if (pred head)
+                 head
+                 (loop (cdr lst)))))))))
+
 (define (launch-service name proc args)
   "Try to start (with PROC) a service providing NAME; return #f on failure.
 Used by `start' and `enforce'."
@@ -876,6 +902,29 @@ daemon writing FILE is running in a separate PID namespace."
               (try-again)
               (apply throw args)))))))
 
+(define %logging-buffer-size
+  ;; Size of the buffer for each line read by logging fibers.
+  512)
+
+(define (read-line! str port)
+  "This is an interruptible version of the 'read-line!' procedure from (ice-9
+rdelim)."
+  ;; As of Guile 3.0.8, (@ (ice-9 rdelim) read-line!) calls
+  ;; '%read-delimited!', which is in C and thus non-interruptible.
+  (define len
+    (string-length str))
+
+  (let loop ((i 0))
+    (and (< i len)
+         (match (read-char port)
+           ((? eof-object? eof)
+            eof)
+           ((or #\newline #\return)
+            i)
+           (chr
+            (string-set! str i chr)
+            (loop (+ i 1)))))))
+
 (define (%service-file-logger file input)
   "Like 'service-file-logger', but doesn't handle the case in which FILE does
 not exist."
@@ -887,15 +936,22 @@ not exist."
     (lambda ()
       (call-with-port output
         (lambda (output)
+          (define line
+            (make-string %logging-buffer-size))
+
           (let loop ()
-            (match (read-line input)
+            (match (read-line! line input)
               ((? eof-object?)
                (close-port input)
                (close-port output))
-              (line
+              (count
                (let ((prefix (strftime default-logfile-date-format
-                                       (localtime (current-time)))))
-                 (format output "~a~a~%" prefix line)
+                                       (localtime (current-time))))
+                     (count  (or count (string-length line))))
+                 ;; Avoid (ice-9 format) to reduce heap allocations.
+                 (put-string output prefix)
+                 (put-string output line 0 count)
+                 (newline output)
                  (loop))))))))))
 
 (define (service-file-logger file input)
@@ -915,18 +971,24 @@ FILE."
   "Return a thunk meant to run as a fiber that reads from INPUT and logs to
 'log-output-port'."
   (lambda ()
+    (define line
+      (make-string %logging-buffer-size))
+
     (let loop ()
-      (match (read-line input)
+      (match (read-line! line input)
         ((? eof-object?)
          (close-port input))
-        (line
+        (count
          (let ((prefix (strftime (%current-logfile-date-format)
-                                 (localtime (current-time)))))
+                                 (localtime (current-time))))
+               (count  (or count (string-length line))))
            ;; TODO: Print the PID of COMMAND.  The actual PID is potentially
            ;; not known until after 'read-pid-file' has completed, so it would
            ;; need to be communicated.
-           (format (log-output-port) "~a[~a] ~a~%"
-                   prefix command line))
+           (simple-format (log-output-port) "~a[~a] "
+                          prefix command)
+           (put-string (log-output-port) line 0 count)
+           (newline (log-output-port)))
          (loop))))))
 
 (define (format-supplementary-groups supplementary-groups)
@@ -1208,9 +1270,13 @@ start."
              pid))
           pid))))
 
-(define* (make-kill-destructor #:optional (signal SIGTERM))
-  "Return a procedure that sends SIGNAL to the process group of the PID given
-as argument, where SIGNAL defaults to `SIGTERM'."
+(define* (make-kill-destructor #:optional (signal SIGTERM)
+                               #:key (grace-period
+                                      (default-process-termination-grace-period)))
+  "Return a procedure that sends @var{signal} to the process group of the PID
+given as argument, where @var{signal} defaults to @code{SIGTERM}.  If the
+process is still running after @var{grace-period} seconds, send it
+@code{SIGKILL}.  The procedure returns once the process has terminated."
   (lambda (pid . args)
     ;; Kill the whole process group PID belongs to.  Don't assume that PID is
     ;; a process group ID: that's not the case when using #:pid-file, where
@@ -1219,8 +1285,10 @@ as argument, where SIGNAL defaults to `SIGTERM'."
     ;; will still be zero (the Shepherd PGID). In that case, use the PID.
     (let ((pgid (getpgid pid)))
       (if (= (getpgid 0) pgid)
-          (kill pid signal) ;don't kill ourself
-          (kill (- pgid) signal)))
+          (terminate-process pid signal           ;don't kill ourself
+                             #:grace-period grace-period)
+          (terminate-process (- pgid) signal
+                             #:grace-period grace-period)))
     #f))
 
 ;; Produce a constructor that executes a command.
@@ -1749,21 +1817,137 @@ otherwise by updating its state."
        ;; Nothing left to wait for.
        #t)
       ((pid . status)
-       (let ((serv (find-service (lambda (serv)
-                                   (and (enabled? serv)
-                                        (match (service-running-value serv)
-                                          ((? number? pid*)
-                                           (= pid pid*))
-                                          (_ #f)))))))
+       ;; Let the process monitor handle it.
+       (put-message (current-process-monitor)
+                    `(handle-process-termination ,pid ,status))
 
-         ;; SERV can be #f for instance when this code runs just after a
-         ;; service's 'stop' method killed its process and completed.
-         (when serv
-           (handle-service-termination serv status))
+       ;; As noted in libc's manual (info "(libc) Process Completion"),
+       ;; loop so we don't miss any terminated child process.
+       (loop)))))
 
-         ;; As noted in libc's manual (info "(libc) Process Completion"),
-         ;; loop so we don't miss any terminated child process.
-         (loop))))))
+(define (process-monitor channel)
+  "Run a process monitor that handles requests received over @var{channel}."
+  (let loop ((waiters vlist-null))
+    (match (get-message channel)
+      (('handle-process-termination pid status)
+       ;; Handle the termination of PID.
+       (match (find-service (lambda (serv)
+                              (and (enabled? serv)
+                                   (match (service-running-value serv)
+                                     ((? number? pid*)
+                                      (= pid pid*))
+                                     (_ #f)))))
+         (#f
+          ;; SERV can be #f for instance when this code runs just after a
+          ;; service's 'stop' method killed its process and completed.
+          #f)
+         ((? service? service)
+          (handle-service-termination service status)))
+
+       ;; Notify any waiters.
+       (vhash-foldv* (lambda (waiter _)
+                       (put-message waiter status)
+                       #t)
+                     #t pid waiters)
+
+       ;; XXX: The call below is linear in the size of WAITERS, but WAITERS is
+       ;; usually empty or small.
+       (loop (vhash-fold (lambda (key value result)
+                           (if (= key pid)
+                               result
+                               (vhash-consv key value result)))
+                         vlist-null
+                         waiters)))
+
+      (('spawn command reply)
+       ;; Spawn COMMAND; send its exit status to REPLY when it terminates.
+       ;; This operation is atomic: the WAITERS table is updated before
+       ;; termination of PID can possibly be handled.
+       (let ((pid (fork+exec-command command)))
+         (loop (vhash-consv pid reply waiters))))
+
+      (('await pid reply)
+       ;; Await the termination of PID and send its status on REPLY.
+       (if (catch-system-error (kill pid 0))
+           (loop (vhash-consv pid reply waiters))
+           (begin                                 ;PID is gone
+             (put-message reply 0)
+             (loop waiters)))))))
+
+(define (spawn-process-monitor)
+  "Spawn a process monitoring fiber and return a channel to communicate with
+it."
+  (define channel
+    (make-channel))
+
+  (spawn-fiber
+   (lambda ()
+     (process-monitor channel)))
+
+  channel)
+
+(define current-process-monitor
+  ;; Channel to communicate with the process monitoring fiber.
+  (make-parameter #f))
+
+(define (call-with-process-monitor thunk)
+  (parameterize ((current-process-monitor (spawn-process-monitor)))
+    (thunk)))
+
+(define-syntax-rule (with-process-monitor exp ...)
+  "Spawn a process monitoring fiber and evaluate @var{exp}... within that
+context.  The process monitoring fiber is responsible for handling
+@code{SIGCHLD} and generally dealing with process creation and termination."
+  (call-with-process-monitor (lambda () exp ...)))
+
+(define (spawn-command program . arguments)
+  "Like 'system*' but do not block while waiting for PROGRAM to terminate."
+  (if (current-process-monitor)
+      (let ((reply (make-channel)))
+        (put-message (current-process-monitor)
+                     `(spawn ,(cons program arguments)
+                             ,reply))
+        (get-message reply))
+      (apply system* program arguments)))
+
+(define default-process-termination-grace-period
+  ;; Default process termination "grace period" before we send SIGKILL.
+  (make-parameter 5))
+
+(define* (get-message* channel timeout #:optional default)
+  "Receive a message from @var{channel} and return it, or, if the message hasn't
+arrived before @var{timeout} seconds, return @var{default}."
+  (call-with-values
+      (lambda ()
+        (perform-operation
+         (choice-operation (get-operation channel)
+                           (sleep-operation timeout))))
+    (match-lambda*
+      (()                               ;'sleep' operation returns zero values
+       default)
+      ((message)                            ;'get' operation returns one value
+       message))))
+
+(define* (terminate-process pid signal
+                            #:key (grace-period
+                                   (default-process-termination-grace-period)))
+  "Send @var{signal} to @var{pid}, which can be negative to denote a process
+group; wait for @var{pid} to terminate and return its exit status.  If
+@var{pid} is still running @var{grace-period} seconds after @var{signal} has
+been sent, send it @code{SIGKILL}."
+  (let ((reply (make-channel)))
+    (put-message (current-process-monitor) `(await ,(abs pid) ,reply))
+    (kill pid signal)
+
+    (match (get-message* reply grace-period #f)
+      (#f
+       (local-output
+        (l10n "Grace period of ~a seconds is over; sending ~a SIGKILL.")
+        grace-period pid)
+       (catch-system-error (kill pid SIGKILL))
+       (get-message reply))
+      (status
+       status))))
 
 (define (handle-service-termination service status)
   "Handle the termination of the process associated with @var{service}, whose
@@ -1930,14 +2114,18 @@ requested to be removed."
   "Shut down all the currently running services; update the persistent state
 file when persistence is enabled."
   (let ((running-services '()))
-    (for-each-service
+    ;; Note: Do not use 'for-each-service' since it introduces a continuation
+    ;; barrier via 'hash-fold', thereby preventing the 'stop' method from
+    ;; suspending via (@ (fibers) sleep), 'spawn-command', or similar.
+    (for-each
      (lambda (service)
        (when (running? service)
          (stop service)
          (when persistency
            (set! running-services
                  (cons (canonical-name service)
-                       running-services))))))
+                       running-services)))))
+     (service-list))
 
     (when persistency
       (call-with-output-file persistency-state-file
